/* loader.c
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "gplbasic.h"
#include "symtab.h"
#include "errors.h"
#include "opcodes.h"
#include "token.h"
#include "strings.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

int used_strstack; /* flag to indicate that string stack was used */

int charareas;
unsigned char *charconst;
int chars_used; /* string constant area */
int floatareas;
double *fltconst; 
int floats_used; /* FP constant area */
static CELLTYPE * codeptr;
int progareas;
int cells_used; /* program memory area */
int memareas;

CELLTYPE * datastart; /* pointer to first data area */
CELLTYPE * cur_data; /* pointer to current data area */
unsigned char defpromptstr[4]={2,128,'?',' '};

static int default_type[26]; /* default type for variables */
static int insub; /* Are we in subroutine? 0=no 1=sub 2=function 3=deffn */
static int parm_slots; /* stack slots on main parm stack */
static int fparm_slots; /* stack slots on FP parm stack */
static int ilocals; 
static int slocals; 
static int flocals; /* number of local variables of each type */
VARSYM *funcvar; /* pseudo variable to return function result */
CELLTYPE *linkptr; /* pointer to OP_LINK instruction at start of routine */

static VARSYM * lpstack[128]; /* local and parameter symbol stack */
int lpptr=0;
#define lppush(p) lpstack[lpptr++]=p
#define lppop(p) p=lpstack[--lpptr]

void clear_program(void) 
{
  int i;
  clear_vars(); /* get rid of all arrays and strings. */
  for(i=0;i<memareas;i++) {
    free(mem_areas[i]); /* Clear all reserved areas for program & symtab */ 
  }
  memareas=0;
  progareas=charareas=floatareas=symareas=labareas=0;
  clear_symtabs();
  progstart=0;
  charconst=0;
  fltconst=0;
  datastart=0;
}

void parse_line(char *);
void parse_statement(void);
static int expr1(void); 
static int expr2(void);
static int expr3(void);
static int expr4(void);
static int expr5(void);
static int expr6(void);
static int expr7(void);
static int expr8(void);
static int expr9(void);
static int expr10(void);
static int expr11(void);
static int expr12(void);
static int expr13(void);
static void mustbeint(int);
static void mustbestr(int);
static void mustbeflt(int);
static int equalize(int,int);
static int parse_factor(void);
static int parse_variable(VARSYM **);
static void parse_defset(int);
static void put_variable(int,VARSYM *);
static void resolve_label(void);
static void use_label(void);
static void use_sub(CELLTYPE *);
static void resolve_sub(CELLTYPE *);
static VARSYM *parse_decl(int);
static void parse_actualparms(CELLTYPE *);
static void close_sub(void);

static void opcode(int);
static void opcint(long);
static void opcptr(void *);

static void newprogmem(void)
     /* Reserve memory area for program */
{
  CELLTYPE *p;
  if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
  else if( (p=malloc(S_PROGCHUNK * sizeof(CELLTYPE))) == 0) {
    doerror(GE_MEM,0,0);
  } else {
    if(progstart==0) {  /* first program area */
      progstart=p;
      codeptr=p;
    } else {
      codeptr->l = OP_JMP;   
      codeptr++;
      codeptr->p = p; /* assemble a jump to the new program area */
      codeptr = p;
    }
    mem_areas[memareas++]=p;
    progareas++;
    cells_used=0;
  }
}

void opcint(long i)
     /* Add integer operand to opcode */
{
  codeptr->l = i;
  codeptr++;
  cells_used++;
}

void opcptr(void * p)
     /* add pointer operand to opcode */
{
  codeptr->p = p;
  codeptr++;
  cells_used++;
}

void opcode(int i)
     /* add opcode to compiled program */
{
  if( progstart==0 || cells_used >= S_PROGCHUNK-4) newprogmem();
  codeptr->l=i;
  codeptr++;
  cells_used++;
}

void bopcode(int i,int s)
     /* add opcode to compiled program with size s*/
{
  if( progstart==0 || cells_used >= S_PROGCHUNK-2-s) newprogmem();
  codeptr->l=i;
  codeptr++;
  cells_used++;
}

static CELLTYPE *doexit; /* linked list of exit points for do loops */

char progline[MAXLINELENGTH+2];
long lineno;	
	
int load_program(char * progname) 
{
  FILE * progfile;
  int i;
  if( memareas != 0 ) clear_program(); /* Nuke previous program if present */
  stkptr=main_stack+S_MAINSTACK;
  lineno=1;
  funcvar=0;parm_slots=fparm_slots=flocals=slocals=ilocals=0;
  insub=0;lpptr=0;
  for(i=0;i<26;i++)default_type[i]=TY_SINGLE;
  if( (progfile = fopen(progname,"r")) == 0) {
    doerror(GE_OFR,progname,0);
  } else {
    newprogmem();
    stkptr--;stkptr->l=KW_END; /* End marker on stack */
    while (fgets( progline, MAXLINELENGTH+2, progfile)) {
      if(lineno==1 && progline[0]=='#' && progline[1]=='!') {
	lineno++;continue;
      }
      parse_line(progline);
      lineno++;
    }
    fclose(progfile);
    if(stkptr->l==KW_FOR) doerror(GE_WIT,"FOR","NEXT");
    opcode(OP_EXIT);
  }
  lineno=0;
  check_unresolved();
  clear_vars();
  stkptr--;stkptr->l=KW_END; /* End marker on stack */
  return 0;
}

void parse_line(char * pp)
{
 parseptr=pp;
 nexttoken();
 nofloat=1;
 used_strstack=0;
 if( thistoken==TOK_INT ) { /* Integer label 'line number' found */
   resolve_label();
   nexttoken();
 } else if ( thistoken==TOK_IDENT && *parseptr==':' ) { 
   /* alphanumeric label found */
   resolve_label();
   nexttoken();
   nexttoken();
 }
 while (thistoken != TOK_EOL) {
   parse_statement();
   if( thistoken==TOK_COLON) nexttoken();
   else if (thistoken!= TOK_EOL) doerror(GE_SYN,0,0);
 }
}


#define ipop(n) do{t=*stkptr++;n=t.l;}while(0)
#define ipush(n) do{if(stkptr==main_stack) doerror(GE_STK,0,0); else\
  {t.l=n;*--stkptr=t;}}while(0)
#define ppop(n) do{t=*stkptr++;n=t.p;}while(0)
#define ppush(n) do{if(stkptr==main_stack) doerror(GE_STK,0,0); else\
  {t.p=n;*--stkptr=t;}}while(0)

void parse_statement(void) 
{
  VARSYM *vp,*vp2;
  int x,nlflag,endflag,needor,usedfile;
  CELLTYPE *p1,*p2;
  CELLTYPE t;
  long a;
  nofloat=0;usedfile=0;
  if(errhandling) {
    bopcode(OP_MARKLINE,4);
    opcint(lineno);
    opcptr(0); /* pointer to last label (for ERL) */
    opcptr(0); /* pointer to next statement for resume next */
  }
  switch(thistoken) {
  case KW_LET:
    nexttoken();
    if(thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
    goto assignment;
  case TOK_IDENT :  /* assignments */
    skipblanks();
    if(*parseptr=='='||*parseptr=='(') { /* assignment */
    assignment: a=parse_variable(&vp);
      require(TOK_EQ); 
      x=expr1();
      if((a&7)==TY_INT) {
        mustbeint(x);
        put_variable(a,vp);
      } else if((a&7)==TY_STR) {
        mustbestr(x);
        put_variable(a,vp);
      } else if((a&7)==TY_DOUBLE) {
        mustbeflt(x);
        put_variable(a,vp);
      }
    } else { /* subroutine call */
      if(vartype!=TY_PLAIN) doerror(GE_SYN,0,0);
      vartype=TY_FUNC+TY_VOID;
      vp=look_var();
      p1=vp->contents.p;
      if(vp->type != TY_FUNC+TY_VOID) doerror(GE_SBR,vp->name,0);
      nexttoken();
      if(p1[0].l & 255) {
        parse_actualparms(p1); 
      }  
      opcode(OP_CALL);
      use_sub(p1);
    }
  break;
  case KW_REM: thistoken=TOK_EOL; break;
  case KW_PRINT: nexttoken();
    nlflag=1;endflag=0;
    if(thistoken==TOK_HASH) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      require(TOK_COMMA);
      opcode(OP_SETOUTPUT);
      usedfile=1;
    }
    do{
      switch(thistoken) {
      case TOK_COMMA: nexttoken();opcode(OP_PRINTC);nlflag=0;break;
      case TOK_SEMI: nexttoken();nlflag=0;break;
      case KW_TAB: nexttoken();x=parse_factor();mustbeint(x);
	opcode(OP_PRINTT);break;
      case TOK_EOL: case TOK_COLON: case KW_ELSE: endflag=1;break;  
      default:
        switch(expr1()) {
        case TY_DOUBLE:
	  opcode(OP_FPRINT);
        break;
        case TY_INT:
          opcode(OP_IPRINT);
        break;
        case TY_STR:
          opcode(OP_SPRINT);
        break;
        }
        nlflag=1;
      }
    }while(!endflag);
    if(nlflag)opcode(OP_PRINTNL);
    if(usedfile) {
      opcode(OP_ILIT);opcint(0);opcode(OP_SETOUTPUT);
    }
  break;
  case KW_WRITE: nexttoken();
    if(thistoken==TOK_HASH) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      require(TOK_COMMA);
      opcode(OP_SETOUTPUT);
      usedfile=1;
    }
    for(;;) {
      switch(expr1()) {
      case TY_INT: opcode(OP_IPRINT);break;
      case TY_DOUBLE: opcode(OP_FPRINT);break;
      case TY_STR: opcode(OP_SWRITE);break;
      }
      if(thistoken==TOK_COMMA) {
	nexttoken();
	opcode(OP_SLIT);opcptr(short_strings+2+3*',');opcode(OP_SPRINT);
      } else break;
    }
    opcode(OP_PRINTNL);
    if(usedfile) {
      opcode(OP_ILIT);opcint(0);opcode(OP_SETOUTPUT);
    }
  break;
  case KW_IF: nexttoken();
    x=expr1();
    mustbeint(x);
    nofloat=1;
    require(KW_THEN);
    if (thistoken==TOK_EOL) { /* multiline if-then statement */
      opcode(OP_BF);
      ppush(codeptr);
      opcptr(0);
      ipush(KW_THEN);
    } else if(thistoken==TOK_INT) { /* is THEN followed by line number? */
      opcode(OP_BT);
      use_label();
      if(thistoken==KW_ELSE) {
        nexttoken();
	opcode(OP_JMP);
	use_label();
      }      
    } else {
      opcode(OP_BF);
      p1=codeptr;
      opcptr(0);
      ipush(KW_IF);
      while (thistoken != TOK_EOL && thistoken != KW_ELSE) {
	parse_statement();
	if( thistoken==TOK_COLON) nexttoken();
	else if (thistoken!= TOK_EOL && thistoken!=KW_ELSE) 
             doerror(GE_SYN,0,0);
      }
      if(thistoken==KW_ELSE) {
        nexttoken();
	opcode(OP_JMP);
	opcptr(0);
	p1->p=codeptr;
	p1=codeptr-1;
	while (thistoken != TOK_EOL) {
	  parse_statement();
	  if( thistoken==TOK_COLON) nexttoken();
	  else if (thistoken!= TOK_EOL) doerror(GE_SYN,0,0);
	}	
      }
      p1->p=codeptr;
      ipop(a);if(a!=KW_IF)doerror(GE_SYN,0,0); 
    }
  break;
  case KW_GOTO: nofloat=1;nexttoken();
    opcode(OP_JMP);
    use_label();
  break;
  case KW_STOP: nexttoken();
    opcode(OP_STOP);
  break;
  case KW_SYSTEM: nexttoken();
    opcode(OP_EXIT);
  break;
  case KW_CLEAR: nexttoken();
    opcode(OP_CLR);
  break;
  case KW_RUN: nexttoken();
    opcode(OP_CLR);
    opcode(OP_JMP);
    opcptr(progstart);
  break;
  case KW_END: nexttoken();
    if(thistoken==KW_IF) goto lab_endif; /* end if */
    else if(thistoken==KW_SUB) {
      nexttoken();
      ipop(a);
      if(insub!=1||a!=KW_CALL) doerror(GE_WIT,"END SUB","SUB");
      ppop(p1);
      close_sub();
      p1->p=codeptr;
    } else if(thistoken==KW_FUNCTION) {
      nexttoken();
      ipop(a);
      if(insub!=2||a!=KW_CALL) doerror(GE_WIT,"END FUNCTION","FUNCTION");
      ppop(p1);
      close_sub();
      p1->p=codeptr;
    } else if(thistoken==KW_DEF) {
      nexttoken();
      ipop(a);
      if(insub!=3||a!=KW_CALL) doerror(GE_WIT,"END DEF","DEF");
      ppop(p1);
      close_sub();
      p1->p=codeptr;
    } else if(thistoken==KW_SELECT) {
      nexttoken();
      ipop(a);
      if(a==KW_IS) { /* select with case else */
	ppop(p2);
	while(p2!=0) {
	  p1=p2->p;
	  p2->p=codeptr;
	  p2=p1; /* resolve reference chain */
	}
      } else if(a==KW_SELECT) { /* without case else */
	ppop(p1);
	ppop(p2);
	ipop(a);
	if(p1!=0) {
	  opcode(OP_JMP);opcptr(p2);p2=codeptr-1;
	  p1->p=codeptr;
	}
	if(a==TY_INT||a==TY_STR) opcode(OP_DROP); else opcode(OP_FDROP);
	while(p2!=0) {
	  p1=p2->p;
	  p2->p=codeptr;
	  p2=p1; /* resolve reference chain */
	}	
      } else doerror(GE_WIT,"END SELECT","SELECT");
    } else opcode(OP_EXIT);
  break;
  case KW_ENDIF: lab_endif: nexttoken();
    ipop(a);if(a!=KW_THEN && a!=KW_ELSE) doerror(GE_WIT,"ENDIF","IF");
    ppop(p1);
    while(p1) { /* resolve all forward jumps to end of if statement */
      p2=p1;p1=p1->p;p2->p=codeptr;
    }
  break;
  case KW_ELSEIF: nexttoken();
    ipop(a);if(a!=KW_THEN) doerror(GE_WIT,"ELSEIF","IF");
    opcode(OP_JMP);
    ppop(p1);          
    p2=codeptr;     
    opcptr(p1->p);  /* create forward branch to end of IF statement */
    p1->p=codeptr;  /* resolve previous conditional branch to ELSEIF */
    x=expr1();
    mustbeint(x);
    require(KW_THEN);
    require(TOK_EOL);
    opcode(OP_BF);   /* create new conditional forward branch. */
    ppush(codeptr);
    opcptr(p2);
    ipush(KW_THEN);
  break;
  case KW_ELSE: nexttoken();
    ipop(a);if(a!=KW_THEN) doerror(GE_WIT,"ELSE","IF");
    opcode(OP_JMP);
    ppop(p1);
    opcptr(p1->p);    /* Create forward referencing branch to end of IF */
    p1->p=codeptr;    /* resolve conditional branch past ELSE */
    ppush(codeptr-1);
    ipush(KW_ELSE);
  break;
  case KW_WHILE: nexttoken();
    p1=codeptr;
    x=expr1();
    mustbeint(x);
    opcode(OP_BF);
    p2=codeptr;
    opcptr(0);
    ppush(p1);
    ppush(p2);
    ipush(KW_WHILE);
  break;
  case KW_WEND: nexttoken();
    ipop(a);if(a!=KW_WHILE) doerror(GE_WIT,"WEND","WHILE");
    ppop(p2);
    ppop(p1);
    opcode(OP_JMP);
    opcptr(p1);
    p2->p=codeptr;
  break;
  case KW_DO: nexttoken();
    ppush(doexit);
    ppush(codeptr);
    if(thistoken==KW_WHILE) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_BF);
      doexit=codeptr;
      opcptr(0);
    } else if (thistoken==KW_UNTIL) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_BT);
      doexit=codeptr;
      opcptr(0);
    } else doexit=0;
    ipush(KW_DO);
  break;
  case KW_LOOP: nexttoken();
    ipop(a);if(a!=KW_DO) doerror(GE_WIT,"LOOP","DO");
    ppop(p1);
    if(thistoken==KW_WHILE) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_BT);
      opcptr(p1);
    } else if(thistoken==KW_UNTIL) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_BF);
      opcptr(p1);
    } else {
      opcode(OP_JMP);
      opcptr(p1);
    }
    while(doexit) {
      p1=doexit->p;
      doexit->p=codeptr;
      doexit=p1;        /* resolve loop exit references */
    }
    ppop(doexit);
  break;
  case KW_EXIT: nexttoken(); 
    if (thistoken==KW_FOR) {
      nexttoken();
      opcode(OP_EXITFOR);
    } else if(thistoken==KW_SUB) {
      nexttoken();
      if(insub!=1) doerror(GE_WIT,"EXIT SUB","SUB");
      bopcode(OP_UNLRET,3);
      opcint(parm_slots);
      opcint(fparm_slots);
    } else if(thistoken==KW_FUNCTION) {
      nexttoken();
      if(insub!=2) doerror(GE_WIT,"EXIT FUNCTION","FUNCTION");
      bopcode(OP_UNLRET,3);
      opcint(parm_slots);
      opcint(fparm_slots);
     } else {
      if (thistoken==KW_DO) nexttoken();
      opcode(OP_JMP);
      opcptr(doexit);
      doexit=codeptr-1;
    } 
  break;
  case KW_GOSUB: nofloat=1;nexttoken();
    opcode(OP_JSR);
    use_label();
  break;
  case KW_RETURN: nexttoken();
    opcode(OP_RET);
  break;
  case KW_INPUT: nexttoken();
    if(thistoken==TOK_HASH) {
      nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_SETINPUT);
      require(TOK_COMMA);
      usedfile=1;
    } else {
      if (thistoken==TOK_STR) {
	parse_factor();
	opcode(OP_SPRINT); /* print prompt string */
	if(thistoken !=TOK_COMMA && thistoken != TOK_SEMI) doerror(GE_SYN,0,0);
	if(thistoken==TOK_SEMI) {
	  opcode(OP_SLIT);opcptr(defpromptstr);
	  opcode(OP_SPRINT);
	}
	nexttoken();
      } else {
	opcode(OP_SLIT);opcptr(defpromptstr);
	opcode(OP_SPRINT);
      }
    }
    opcode(OP_INPUTNL);
    for(;;) {
      if (thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
      a=parse_variable(&vp);
      if((a&7)==TY_INT) {
	opcode(OP_IINPUT);
      } else if((a&7)==TY_DOUBLE) {
	opcode(OP_FINPUT);
      } else if((a&7)==TY_STR) {
	opcode(OP_SINPUT);
	used_strstack=1;
      } else doerror(GE_TYP,0,0);
      put_variable(a,vp);
      if(thistoken==TOK_COMMA) {
	nexttoken();
	opcode(OP_INPUTCOM);
      } else break;
    }
    if(usedfile) {
      opcode(OP_ILIT);opcint(0);opcode(OP_SETINPUT);
    }
  break;
  case KW_LINE: nexttoken();
    if(thistoken==KW_INPUT) {
      nexttoken(); 
      if(thistoken==TOK_HASH) {
	nexttoken();
	x=expr1();
	mustbeint(x);
	require(TOK_COMMA);
	opcode(OP_SETINPUT);
	usedfile=1;
      } else {
	if(thistoken==TOK_STR) {
	  x=parse_factor();
	  opcode(OP_SPRINT);
	  require(TOK_SEMI);
	}
      }
      opcode(OP_INPUTNL);
      if(thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
      a=parse_variable(&vp);
      if((a&7)!=TY_STR) doerror(GE_TYP,0,0);
      opcode(OP_SINPUTL);
      used_strstack=1;
      put_variable(a,vp);
      if(usedfile) {
	opcode(OP_ILIT);opcint(0);opcode(OP_SETINPUT);
      }
    } else doerror(GE_SYN,0,0);
  break;
  case KW_FOR: nexttoken();
    if (thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
    parse_variable(&vp);
    if(vp->type==TY_INT||vp->type==TY_LONG) {
      require(TOK_EQ);
      x=expr1();
      mustbeint(x);
      require(KW_TO);
      x=expr1();
      mustbeint(x);
      if(thistoken==KW_STEP) { nexttoken();
       x=expr1();
       mustbeint(x);
      } else {
        opcode(OP_ILIT);opcint(1);
      }
      if(vp->contents.l==1||(vp->contents.l==0 &&(insub==0||insub==3))) {
	bopcode(OP_IFOR,3);
	opcptr(&(vp->contents.l));
      } else if (vp->contents.l<=-2) {
	bopcode(OP_IFORL,3);opcint(vp->contents.l+1);
      } else if (vp->contents.l==0) {
	vp->contents.l=-2-(ilocals++);
	lppush(vp);
	bopcode(OP_IFORL,3);opcint(vp->contents.l+1);
      } else doerror(GE_TYP,0,0);
      opcptr(0);
    } else if(vp->type==TY_DOUBLE||vp->type==TY_SINGLE) {
      require(TOK_EQ);
      x=expr1();
      mustbeflt(x);
      require(KW_TO);
      x=expr1();
      mustbeflt(x);
      if(thistoken==KW_STEP) { nexttoken();
       x=expr1();
       mustbeflt(x);
      } else {
        opcode(OP_ILIT);opcint(1);opcode(OP_I2F);
      }
      if(vp->contents.l==1||(vp->contents.l==0 &&(insub==0||insub==3))) {
	bopcode(OP_FFOR,3);
	opcptr(&(vp->contents.d));
      } else if (vp->contents.l<=-2) {
	bopcode(OP_FFORL,3);opcint(vp->contents.l+1);
      } else if (vp->contents.l==0) {
	vp->contents.l=-2-(flocals++);
	lppush(vp);
	bopcode(OP_FFORL,3);opcint(vp->contents.l+1);
      } else doerror(GE_TYP,0,0);
      opcptr(0);      
    } else doerror(GE_TYP,0,0);
    ppush(codeptr-1);
    ppush(vp);
    ipush(KW_FOR);
  break;
  case KW_NEXT: nexttoken();
    ipop(a); if(a!=KW_FOR) doerror(GE_WIT,"NEXT","FOR");
    ppop(vp2);
    if(thistoken==TOK_IDENT) { /* check variable */
      parse_variable(&vp);
      if(vp!=vp2) doerror(GE_WIT,"NEXT","FOR");
    }
    if(vp2->type==TY_INT ||vp2->type==TY_LONG) 
      opcode(OP_INEXT);else opcode(OP_FNEXT);
    ppop(p1);
    p1->p=codeptr;  
  break;
  case KW_ON: nexttoken();
    x=expr1();
    mustbeint(x);
    nofloat=1;
    if(thistoken==KW_GOTO) bopcode(OP_TJMP,128);
    else if(thistoken==KW_GOSUB) bopcode(OP_TJSR,128);
    else doerror(GE_SYN,0,0);
    p1=codeptr;a=0;
    opcint(0);
    do {
      nexttoken();
      use_label();
      a++;
    } while(thistoken==TOK_COMMA);
    p1->l=a; /* put number of addresses directly after opcode */
  break;
  case KW_DIM:
    do {
      nexttoken();
      if (thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
      a=parse_variable(&vp);
      if((a&255)==TY_INT+TY_ARRAY) {
	bopcode(OP_IDIM,3);
	opcint(a>>8);
	opcptr(&(vp->contents.p));
      } else if((a&255)==TY_DOUBLE+TY_ARRAY) {
	bopcode(OP_FDIM,3);
	opcint(a>>8);
	opcptr(&(vp->contents.p));
      } else if((a&255)==TY_STR+TY_ARRAY) {
	bopcode(OP_SDIM,3);
	opcint(a>>8);
	opcptr(&(vp->contents.p));
      } else doerror(GE_TYP,0,0);
    } while(thistoken==TOK_COMMA);
  break;
  case KW_RANDOMIZE: nexttoken(); 
    if(thistoken==TOK_COLON || thistoken==KW_ELSE || thistoken==TOK_EOL) {
      opcode(OP_ILIT);opcint(-1);
    } else {
      x=expr1();mustbeint(x);
    }
    opcode(OP_SRAND);
  break;
  case KW_READ: 
    do {
      nexttoken();
      a=parse_variable(&vp);
      opcode(OP_READ);
      if((a&7)==TY_INT) {
	opcode(OP_VAL);
	opcode(OP_F2I);
      } else if((a&7)==TY_DOUBLE) {
	opcode(OP_VAL);
      } else if((a&7)==TY_STR) {
        /* do nothing */
      } else doerror(GE_TYP,0,0);
      put_variable(a,vp);
    } while(thistoken==TOK_COMMA);
  break;
  case KW_RESTORE: nofloat=1;nexttoken();
    opcode(OP_RESTOR);
    if(thistoken==TOK_COLON||thistoken==KW_ELSE||thistoken==TOK_EOL) {
      opcptr(0);
    } else {
      use_label();
    }
  break;
  case KW_DATA: 
    bopcode(OP_SKDATA,128);
    a=0;
    opcptr(0); /* address after data statement */
    opcptr(0); /* address of next data area */
    if(datastart==0) {
      datastart=codeptr-3;cur_data=datastart;
    } else {
      cur_data[2].p=codeptr-3;
      cur_data=codeptr-3;
    }
    do {
      if(*parseptr==',') parseptr++;
      skipblanks();
      if(*parseptr=='"') { /*quoted string */
	nexttoken();
      } else {
	x=0;
	while(*parseptr!='\0' && *parseptr!=',') tokname[x++]=*parseptr++;
	tokname[x]='\0';
	while(x>0 && ( tokname[x-1]==' '||tokname[x-1]=='\t'
	      ||tokname[x-1]=='\n')) {
	  tokname[--x]=0;
	}	
      } /* collect data string in tokname */
      x=strlen(tokname)+2;
      if(x==2) { /* null string */
	opcptr(short_strings);
      } else if (x==3) { /* single character string */
	opcptr(short_strings+2+3*tokname[0]);
      } else {
	if ( chars_used + x > S_STRCONST || charconst == 0) {
	  if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
	  if( !(mem_areas[memareas]=malloc(S_STRCONST))) doerror(GE_MEM,0,0);
	  charconst=mem_areas[memareas++];
	  charareas++;
	  chars_used=0;
	}
	opcptr(charconst);
	memcpy(charconst+2,tokname,x-2);
	*charconst=(x-2)&255;
	*(charconst+1)=128|((x-2)>>8);
	charconst+=x;
	chars_used+=x;
      } /* allocate data string as string constant */
      a++;
      skipblanks();
    } while(*parseptr==',');
    nexttoken();
    cur_data[1].p=codeptr;
  break;
  case KW_DEF: nexttoken();
    if(insub)doerror(GE_SYN,0,0);
    vp=parse_decl(7);
    opcode(OP_JMP); /* jump past function */
    ppush(codeptr); 
    ipush(KW_CALL);
    opcptr(0);
    p1=vp->contents.p;
    resolve_sub(p1);
    insub=3;
    opcode(OP_LINK);linkptr=codeptr;opcint(0);opcint(0);opcint(0);
    if(thistoken==TOK_EQ) { /* it's a single line def */
      nexttoken();
      x=expr1();
      a=(vp->type&7);
      if(a==TY_INT||a==TY_LONG) {
	mustbeint(x);
	opcode(OP_IPUTL);opcint(parm_slots+5);
      } else if(a==TY_DOUBLE||a==TY_SINGLE) {
	mustbeflt(x);
	opcode(OP_FPUTL);opcint(fparm_slots);
      } else {
	mustbestr(x);
	opcode(OP_SPUTL2);opcint(parm_slots+5);
      }
      ipop(a);
      ppop(p1);
      close_sub();
      p1->p=codeptr;
    }
  break;
  case KW_DEFINT:
    parse_defset(TY_INT);
  break;
  case KW_DEFLNG:
    parse_defset(TY_LONG);
  break;
  case KW_DEFSNG:
    parse_defset(TY_SINGLE);
  break;
  case KW_DEFDBL:
    parse_defset(TY_DOUBLE);
  break;
  case KW_DEFSTR:
    parse_defset(TY_STR);
  break;
  case KW_OPTION: nexttoken();
    /* Even the ANSI standard for Minimal BASIC requires OPTION BASE,
       so we provide a dummy. */
       require(KW_BASE);
       require(TOK_INT);
  break;
  case KW_SUB: nexttoken();
    if(insub)doerror(GE_SYN,0,0);
    vp=parse_decl(2);
    if(strcmp(vp->name,"MAIN")==0) {
      bopcode(OP_CALL,4);opcptr(codeptr+3);
    } 
    opcode(OP_JMP); /* jump past subroutine */
    ppush(codeptr);
    ipush(KW_CALL);
    opcptr(0);
    p1=vp->contents.p;
    resolve_sub(p1);
    insub=1;
    opcode(OP_LINK);linkptr=codeptr;opcint(0);opcint(0);opcint(0);
  break;  
  case KW_FUNCTION: nexttoken();
    if(insub)doerror(GE_SYN,0,0);
    vp=parse_decl(3);
    opcode(OP_JMP); /* jump past function */
    ppush(codeptr);
    ipush(KW_CALL);
    opcptr(0);
    p1=vp->contents.p;
    resolve_sub(p1);
    insub=2;
    opcode(OP_LINK);linkptr=codeptr;opcint(0);opcint(0);opcint(0);
  break;  
  case KW_CALL: nexttoken();
    if(thistoken!=TOK_IDENT||vartype!=TY_PLAIN) doerror(GE_SYN,0,0);
    vartype=TY_FUNC+TY_VOID;
    vp=look_var();
    p1=vp->contents.p;
    if(vp->type != TY_FUNC+TY_VOID) doerror(GE_SBR,vp->name,0);
    nexttoken();
    if(p1[0].l & 255) {
      if(thistoken==TOK_LPAR) {
       nexttoken();
       parse_actualparms(p1); 
       require(TOK_RPAR);
      } else parse_actualparms(p1);
    }  
    opcode(OP_CALL);
    use_sub(p1);
  break;
  case KW_DECLARE: nexttoken();
    if(thistoken==KW_SUB) {
      nexttoken();
      parse_decl(0);
    } else if(thistoken==KW_FUNCTION) {
      nexttoken();
      parse_decl(1);
    }
  break;
  case KW_SHARED: 
    do{
      nexttoken();
      parse_variable(&vp);
      if(vp->type&(TY_ARRAY|TY_FUNC)) doerror(GE_TYP,0,0);
      vp->contents.l=1;
      lppush(vp);
    } while(thistoken==TOK_COMMA);
  break;
  case KW_SELECT: nexttoken();require(KW_CASE);
    x=expr1();
    ipush(x);
    ppush(0); /* chain of references out of the select statemment */
    ppush(0); /* reference to the next case clause */
    ipush(KW_SELECT);
    used_strstack=0; /* if the case expr was a string on stack do'nt clear it*/
  break;
  case KW_CASE: nexttoken();
    /* lots of syntactic saccharine here */
    ipop(a);if(a!=KW_SELECT) doerror(GE_WIT,"CASE","SELECT");
    ppop(p1);
    ppop(p2);
    ipop(a);
    if(p1!=0) { /* not the first case in the chain */
      opcode(OP_JMP);opcptr(p2);p2=codeptr-1;
      /* add jump out of select statement to previous case */
      p1->p=codeptr; /* jump from previous case to this one */
    }
    if(thistoken!= KW_ELSE) {
      needor=0;
      if(a==TY_INT||a==TY_STR) opcode(OP_DUP); else opcode(OP_FDUP);
      for(;;) {
	if(thistoken== KW_IF || thistoken ==KW_IS) { 
	  nexttoken();
	  endflag=thistoken;nexttoken();x=expr1();
	  x=equalize(a,x);
	  switch(endflag) {
	  case TOK_LT: 
	    if(x==TY_INT) opcode(OP_ILT); 
	    else if(x==TY_DOUBLE) opcode(OP_FLT); else opcode(OP_SLT);
	    break;
	  case TOK_GE: 
	    if(x==TY_INT) opcode(OP_IGE); 
	    else if(x==TY_DOUBLE) opcode(OP_FGE); else opcode(OP_SGE);
	    break;
	  case TOK_GT: 
	    if(x==TY_INT) opcode(OP_IGT); 
	    else if(x==TY_DOUBLE) opcode(OP_FGT); else opcode(OP_SGT);
	    break;
	  case TOK_LE:
	    if(x==TY_INT) opcode(OP_ILE); 
	    else if(x==TY_DOUBLE) opcode(OP_FLE); else opcode(OP_SLE);
	    break;
	  case TOK_EQ:
	    if(x==TY_INT) opcode(OP_IEQ); 
	    else if(x==TY_DOUBLE) opcode(OP_FEQ); else opcode(OP_SEQ);
	    break;
	  case TOK_NE:
	    if(x==TY_INT) opcode(OP_INE); 
	    else if(x==TY_DOUBLE) opcode(OP_FNE); else opcode(OP_SNE);
	    break;
	  default: doerror(GE_SYN,0,0);
	  }
	} else { /* no IF or IS */
	  x=expr1();
	  x=equalize(a,x);
	  if(thistoken==KW_TO) {
	    if(x==TY_INT) opcode(OP_IGE);
	    else if(x==TY_DOUBLE) opcode(OP_FGE);
	    else opcode(OP_SGE);
	    nexttoken();
	    if(a==TY_INT || a==TY_STR) opcode(needor?OP_2PICK:OP_OVER); 
	                               else opcode(OP_FDUP);
	    /* get the test value again */
	    x=expr1();
	    x=equalize(a,x);
	    if(x==TY_INT) opcode(OP_ILE);
	    else if(x==TY_DOUBLE) opcode(OP_FLE);
	    else opcode(OP_SLE);
	    opcode(OP_IAND);
	  } else { /* a single case expression */
	    if(x==TY_INT) opcode(OP_IEQ);
	    else if(x==TY_DOUBLE) opcode(OP_FEQ);
	    else opcode(OP_SEQ);
	  }
	}
	if(needor) opcode(OP_IOR);
	if(thistoken==TOK_COMMA) { /* yet another case expression */
	  nexttoken();
	  needor=1;
	  if(a==TY_INT || a==TY_STR) opcode(OP_OVER); else opcode(OP_FDUP); 
	} else break;
      }
       /*(runtime)at this point we have a single truth value on the stack */
      opcode(OP_BF);p1=codeptr;opcptr(0); /* jump to next case */
      if(a==TY_INT||a==TY_STR) opcode(OP_DROP); else opcode(OP_FDROP);
      /* get rid of test value once a case is chosen */
        ipush(a);ppush(p2);ppush(p1);ipush(KW_SELECT);
    } else { /* CASE ELSE */
      nexttoken();
      if(a==TY_INT||a==TY_STR) opcode(OP_DROP); else opcode(OP_FDROP);
      ppush(p2);
      ipush(KW_IS);
    }
  break;
  case KW_OPEN: nexttoken();
    x=expr1();
    mustbestr(x);
    if(thistoken==TOK_COMMA) { /* old OPEN syntax */
      nexttoken();
      if(thistoken==TOK_HASH)nexttoken();
      x=expr1();
      mustbeint(x);
      require(TOK_COMMA);
      x=expr1();
      mustbestr(x);
    } else { /* new OPEN syntax */
      require(KW_FOR);
      opcode(OP_SLIT);
      if(thistoken==KW_INPUT) opcptr(short_strings + 2 + 3 * 'I');
      else if (thistoken==KW_OUTPUT) opcptr(short_strings + 2 + 3 * 'O');
      else if (thistoken==KW_APPEND) opcptr(short_strings + 2 + 3 * 'A');
      else doerror(GE_SYN,0,0);
      nexttoken();
      require(KW_AS);
      require(TOK_HASH);
      x=expr1();
      mustbeint(x);
      opcode(OP_ROT); /* move file name argument to top of stack */
    }
    opcode(OP_FOPEN);
  break;
  case KW_CLOSE:
    do {
      nexttoken();
      if(thistoken==TOK_HASH)nexttoken();
      x=expr1();
      mustbeint(x);
      opcode(OP_FCLOSE);
    } while(thistoken==TOK_COMMA);
  break;
  default: doerror(GE_SYN,0,0);
  }
  if(used_strstack) {
    opcode(OP_CLRSTK);
    used_strstack=0;
  }
}

void mustbeint(int x) 
     /* Check that expression is integer */
{
 if(x == TY_DOUBLE) opcode(OP_F2I);
 else if (x != TY_INT) doerror(GE_TYP,0,0);
}

void mustbestr(int x)
{
 if(x != TY_STR) doerror(GE_TYP,0,0);
}

void mustbeflt(int x)
{
  if(x == TY_INT) opcode(OP_I2F);
  else if(x != TY_DOUBLE) doerror(GE_TYP,0,0);
}

/* Now this is sick! Q(UICK)Basic seems to have 13 precedence levels! */

int expr1(void)
{
 int x;
 x=expr2();
 while ( thistoken==KW_IMP ) {
   nexttoken();
   mustbeint(x);
   x=expr2();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IIMP);
 }
 return x; 
}

int expr2(void)
{
 int x;
 x=expr3();
 while ( thistoken==KW_EQV ) {
   nexttoken();
   mustbeint(x);
   x=expr3();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IEQV);
 } 
 return x;
}

int expr3(void)
{
 int x;
 x=expr4();
 while ( thistoken==KW_XOR ) {
   nexttoken();
   mustbeint(x);
   x=expr4();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IXOR);
 } 
 return x;
}

int expr4(void)
{
 int x;
 x=expr5();
 while ( thistoken==KW_OR ) {
   nexttoken();
   mustbeint(x);
   x=expr5();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IOR);
 } 
 return x;
}

int expr5(void)
{
 int x;
 x=expr6();
 while ( thistoken==KW_AND ) {
   nexttoken();
   mustbeint(x);
   x=expr6();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IAND);
 } 
 return x;
}

int expr6(void)
{
 int x;
 if ( thistoken==KW_NOT ) {
  nexttoken();
  x=expr7();
  mustbeint(x);
  x=TY_INT;
  opcode(OP_INOT);
 } else x=expr7();
 return x;
}

static int equalize(int x1,int x2)
{
  /* ensure that operand types are equal */
  if( x1==TY_DOUBLE ) mustbeflt(x2);
  else if( x1==TY_INT ) {
    if (x2==TY_DOUBLE) { /* convert expression 1 to double */
         opcode(OP_I2F2);x1=TY_DOUBLE;
    } else if (x2 != TY_INT) doerror(GE_TYP,0,0);
  } else if (x1 != x2) doerror(GE_TYP,0,0);
  return(x1);
}

int expr7(void)
{
 int x1,x2;
 x1=expr8();
 for(;;) {
   switch(thistoken) {
   case TOK_EQ:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_IEQ);
     else if(x1==TY_DOUBLE) opcode(OP_FEQ);
     else if(x1==TY_STR) opcode(OP_SEQ);
     x1=TY_INT;
   break;
   case TOK_NE:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_INE);
     else if(x1==TY_DOUBLE) opcode(OP_FNE);
     else if(x1==TY_STR) opcode(OP_SNE);
     x1=TY_INT;
   break;
   case TOK_LT:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_ILT);
     else if(x1==TY_DOUBLE) opcode(OP_FLT);
     else if(x1==TY_STR) opcode(OP_SLT);
     x1=TY_INT;
   break;
   case TOK_GE:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_IGE);
     else if(x1==TY_DOUBLE) opcode(OP_FGE);
     else if(x1==TY_STR) opcode(OP_SGE);
     x1=TY_INT;
   break;
   case TOK_GT:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_IGT);
     else if(x1==TY_DOUBLE) opcode(OP_FGT);
     else if(x1==TY_STR) opcode(OP_SGT);
     x1=TY_INT;
   break;
   case TOK_LE:
     nexttoken();
     x2=expr8();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_ILE);
     else if(x1==TY_DOUBLE) opcode(OP_FLE);
     else if(x1==TY_STR) opcode(OP_SLE);
     x1=TY_INT;
   break;
   default: return x1; 
   }
 }
}

static int equalarith(int x1,int x2)
     /* ensure that arithmetic operands are of same type */
{
  if( x1==TY_DOUBLE ) mustbeflt(x2);
  else if( x1==TY_INT ) {
    if (x2==TY_DOUBLE) { /* convert expression 1 to double */
         opcode(OP_I2F2);x1=TY_DOUBLE;
    } else if (x2 != TY_INT) doerror(GE_TYP,0,0);
  } else doerror(GE_TYP,0,0);
  return(x1);
}

int expr8(void)
{
 int x1,x2;
 x1=expr9();
 for(;;) {
   if(thistoken==TOK_PLUS) {
     nexttoken();
     x2=expr9();
     x1=equalize(x1,x2);
     if(x1==TY_INT) opcode(OP_IADD);
     else if(x1==TY_DOUBLE) opcode(OP_FADD);
     else if(x1==TY_STR) {opcode(OP_SCAT);used_strstack=1;}
   } else if(thistoken==TOK_MINUS) {
     nexttoken();
     x2=expr9();
     x1=equalarith(x1,x2);
     if(x1==TY_INT) opcode(OP_ISUB);
     else opcode(OP_FSUB);
   } else return x1;
 }
}

int expr9(void)
{
 int x;
 x=expr10();
 while ( thistoken==KW_MOD ) {
   nexttoken();
   mustbeint(x);
   x=expr10();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IMOD);
 } 
 return x;
}

int expr10(void)
{
 int x;
 x=expr11();
 while ( thistoken==TOK_IDIV ) {
   nexttoken();
   mustbeint(x);
   x=expr11();
   mustbeint(x);
   x=TY_INT;
   opcode(OP_IDIV);
 } 
 return x;
}

int expr11(void)
{
 int x1,x2;
 x1=expr12();
 for(;;) {
   if(thistoken==TOK_TIMES) {
     nexttoken();
     x2=expr12();
     x1=equalarith(x1,x2);
     if(x1==TY_INT) opcode(OP_IMUL);
     else opcode(OP_FMUL);
   } else if(thistoken==TOK_DIV) {
     /* / operator is always floating point */
     mustbeflt(x1);
     nexttoken(); 
     x2=expr12();
     mustbeflt(x2);
     opcode(OP_FDIV);
     x1=TY_DOUBLE;
   } else return x1;
 }
}

int expr12(void)
{
 int x;
 if ( thistoken==TOK_MINUS ) {
  nexttoken();
  x=expr13();
  if(x==TY_INT) opcode(OP_INEG);
  else if(x==TY_DOUBLE) opcode(OP_FNEG);
  else doerror(GE_TYP,0,0);  
 } else x=expr13();
 return x;
}


int expr13(void)
{ 
  /* this is for the exponentiation operator */
 int x;
 x=parse_factor();
 while(thistoken==TOK_POW) {
   nexttoken();
   mustbeflt(x);
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_FFPOW);
   x=TY_DOUBLE;
 }
 return x;
}

int parse_factor() 
{
 int x;
 VARSYM *vp;
 long l;
 double d;
 switch( thistoken) {
 case TOK_LPAR: 
   nexttoken();
   x=expr1();
   require(TOK_RPAR);
 break;
 case TOK_INT:
   l=atol(tokname);
   opcode(OP_ILIT);
   opcint(l);
   nexttoken();
   x=TY_INT;
 break;
 case TOK_OCT:
   l=strtoul(tokname,0,8);
   opcode(OP_ILIT);
   opcint(l);
   nexttoken();
   x=TY_INT;
 break;
 case TOK_HEX:
   l=strtoul(tokname,0,16);
   opcode(OP_ILIT);
   opcint(l);
   nexttoken();
   x=TY_INT;
 break;
 case TOK_FLT:
   d=atof(tokname);
   opcode(OP_FGET);
   if ( floats_used > S_FCONST || fltconst == 0) {
     if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
     if( !(mem_areas[memareas]=malloc(S_FCONST*sizeof(double)))) 
                         doerror(GE_MEM,0,0);
     fltconst=mem_areas[memareas++];
     floatareas++;
     floats_used=0;
   }
   opcptr(fltconst);
   *fltconst++=d;
   floats_used++;
   nexttoken();
   x=TY_DOUBLE;  
 break;
 case TOK_IDENT:
   x=parse_variable(&vp);
   if(x==TY_INT) {
     if(vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
       opcode(OP_IGET); /* global */
       opcptr( &(vp->contents.l));
     } else if (vp->contents.l>=2) { /* parameter */
       opcode(OP_IGETR);
       opcint(parm_slots-vp->contents.l+6);
     } else if (vp->contents.l<=-2) { /* existing local */
       opcode(OP_IGETL);opcint(vp->contents.l+1);
     } else if (vp->contents.l==0) { /* new local */
       vp->contents.l=-2-(ilocals++);
       opcode(OP_IGETL);opcint(vp->contents.l+1);
       lppush(vp);
     }
   } else if(x==TY_STR) {
     if(vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
       opcode(OP_SGET);
       opcptr( &(vp->contents.p));
     } else if (vp->contents.l>=2) {
       opcode(OP_SGETR);
       opcint(parm_slots-vp->contents.l+6);
     } else if (vp->contents.l<=-2) {
       opcode(OP_SGETL);opcint(vp->contents.l+1);
     } else if (vp->contents.l==0) {
       vp->contents.l=-2-(slocals++);
       opcode(OP_SGETL);opcint(vp->contents.l+1);
       lppush(vp);
     }
   } else if(x==TY_DOUBLE) {
     if (vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
       opcode(OP_FGET);
       opcptr( &(vp->contents.d));
     } else if (vp->contents.l>=2) {
       opcode(OP_FGETR);
       opcint(parm_slots-vp->contents.l+6);
     } else if (vp->contents.l<=-2) {
       opcode(OP_FGETL);opcint(vp->contents.l+1);
     } else if (vp->contents.l==0) {
       vp->contents.l=-2-(flocals++);
       opcode(OP_FGETL);opcint(vp->contents.l+1);
       lppush(vp);
     }
   } else if ((x&255)==TY_INT+TY_ARRAY) {
     bopcode(OP_IGETA,3);
     opcint(x>>8);
     opcptr( &(vp->contents.p));
     x=TY_INT;
   } else if ((x&255)==TY_STR+TY_ARRAY) {
     bopcode(OP_SGETA,3);
     opcint(x>>8);
     opcptr( &(vp->contents.p));
     x=TY_STR;
   } else if ((x&255)==TY_DOUBLE+TY_ARRAY) {
     bopcode(OP_FGETA,3);
     opcint(x>>8);
     opcptr( &(vp->contents.p));
     x=TY_DOUBLE;
   } else if (x&TY_FUNC) { /* it's a function */
     /* reserve room for function result */
     if((x&7)==TY_INT || (x&7)==TY_LONG) {
       opcode(OP_ILIT);opcint(0);x=TY_INT;
     } else if ((x&7)==TY_DOUBLE || (x&7)==TY_SINGLE) {
       opcode(OP_FZERO);x=TY_DOUBLE;
     } else {
       opcode(OP_SLIT);opcptr(short_strings);x=TY_STR;
     }
     nexttoken();         
     if(vp->type&64) {
       require(TOK_LPAR);
       parse_actualparms(vp->contents.p);
       require(TOK_RPAR);
     }
     opcode(OP_CALL);
     use_sub(vp->contents.p);
     used_strstack=1;
   } else doerror(GE_TYP,0,0);
 break;
 case TOK_STR: /* string constant */
   l=strlen(tokname)+2;
   if(l==2) { /* null string */
     opcode(OP_SLIT);
     opcptr(short_strings);
   } else if (l==3) { /* single character string */
     opcode(OP_SLIT);
     opcptr(short_strings+2+3*tokname[0]);
   } else {
     if ( chars_used + l > S_STRCONST || charconst == 0) {
       if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
       if( !(mem_areas[memareas]=malloc(S_STRCONST))) doerror(GE_MEM,0,0);
       charconst=mem_areas[memareas++];
       charareas++;
       chars_used=0;
     }
     opcode(OP_SLIT);
     opcptr(charconst);
     memcpy(charconst+2,tokname,l-2);
     *charconst=(l-2)&255;
     *(charconst+1)=128|((l-2)>>8);
     charconst+=l;
     chars_used+=l;
   }
   nexttoken();
   x=TY_STR;
 break;
 case KW_CHR:nexttoken(); 
   x=parse_factor();
   mustbeint(x);
   opcode(OP_CHR);
   x=TY_STR;
 break;
 case KW_LEN:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_LEN);
   x=TY_INT;
 break;
 case KW_ASC:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_ASC);
   x=TY_INT;
 break;
 case KW_ABS:nexttoken();
   x=parse_factor();
   if(x==TY_INT) opcode(OP_IABS);
   else if(x==TY_DOUBLE) opcode(OP_FABS);
   else doerror(GE_TYP,0,0);
 break;
 case KW_SGN:nexttoken();
   x=parse_factor();
   if (x==TY_INT) opcode(OP_ISGN);
   else if(x==TY_DOUBLE) opcode(OP_FSGN);
   else doerror(GE_TYP,0,0);
   x=TY_INT;
 break;
 case KW_SIN:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_SIN);
   x=TY_DOUBLE;
 break;
 case KW_COS:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_COS);
   x=TY_DOUBLE;
 break;
 case KW_TAN:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_TAN);
   x=TY_DOUBLE;
 break;
 case KW_ATN:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_ATN);
   x=TY_DOUBLE;
 break;
 case KW_SQR:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_SQR);
   x=TY_DOUBLE;
 break;
 case KW_LOG:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_LOG);
   x=TY_DOUBLE;
 break;
 case KW_EXP:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_EXP);
   x=TY_DOUBLE;
 break;
 case KW_RND:nexttoken();
   if(thistoken==TOK_LPAR) {
     nexttoken();
     x=expr1();
     mustbeflt(x);
     require(TOK_RPAR);
   } else {
     opcode(OP_ILIT);opcint(1);opcode(OP_I2F);
   }
   opcode(OP_RND);
   x=TY_DOUBLE;
 break;
 case KW_INT:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   opcode(OP_INT);
   x=TY_DOUBLE;
 break;
 case KW_LEFT:nexttoken();
   require(TOK_LPAR);
   x=expr1();
   mustbestr(x);
   require(TOK_COMMA);
   x=expr1();
   mustbeint(x);
   require(TOK_RPAR);
   opcode(OP_LEFT);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_RIGHT:nexttoken();
   require(TOK_LPAR);
   x=expr1();
   mustbestr(x);
   require(TOK_COMMA);
   x=expr1();
   mustbeint(x);
   require(TOK_RPAR);
   opcode(OP_RIGHT);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_MID:nexttoken();
   require(TOK_LPAR);
   x=expr1();
   mustbestr(x);
   require(TOK_COMMA);
   x=expr1();
   mustbeint(x);
   if(thistoken==TOK_COMMA) {
     nexttoken();
     x=expr1();
     mustbeint(x);
     opcode(OP_MID);
   } else opcode(OP_MID2);
   require(TOK_RPAR);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_STR:nexttoken();
   x=parse_factor();
   mustbeflt(x);
   used_strstack=1;
   opcode(OP_STR);
   x=TY_STR;
 break;
 case KW_INSTR: nexttoken();
   require(TOK_LPAR);
   x=expr1();
   if(x==TY_STR) { /* no first numeric expression */
     opcode(OP_ILIT);
     opcint(1);
     opcode(OP_SWAP);
   } else {
     mustbeint(x);
     require(TOK_COMMA);
     x=expr1();
   }
   mustbestr(x);
   require(TOK_COMMA);
   x=expr1();
   mustbestr(x);
   require(TOK_RPAR);
   opcode(OP_INSTR);
   x=TY_INT;
 break;
 case KW_STRING:nexttoken();
   require(TOK_LPAR);
   x=expr1();
   mustbeint(x);
   require(TOK_COMMA);
   x=expr1();
   if(x==TY_STR) {
     opcode(OP_ASC);
   } else {
     mustbeint(x);
   }
   opcode(OP_STRING);
   require(TOK_RPAR);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_VAL:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_VAL);
   x=TY_DOUBLE;
 break;
 case KW_TIME:nexttoken();
   opcode(OP_TIME);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_DATE:nexttoken();
   opcode(OP_DATE);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_TIMER:nexttoken();
   opcode(OP_TIMER);
   x=TY_INT;
 break;
 case KW_OCT:nexttoken();
   x=parse_factor();
   mustbeint(x);
   opcode(OP_OCT);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_HEX:nexttoken();
   x=parse_factor();
   mustbeint(x);
   opcode(OP_HEX);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_UCASE:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_UCASE);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_LCASE:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_LCASE);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_LTRIM:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_LTRIM);
   used_strstack=1;
   x=TY_STR;
 break;
 case KW_RTRIM:nexttoken();
   x=parse_factor();
   mustbestr(x);
   opcode(OP_RTRIM);
   used_strstack=1;
   x=TY_STR;
 break;   
 case KW_CINT: case KW_CLNG: case KW_FIX: nexttoken();
   x=parse_factor();
   mustbeint(x);
   x=TY_INT;
 break;
 case KW_CSNG: case KW_CDBL: nexttoken();
   x=parse_factor();
   mustbeflt(x);
   x=TY_DOUBLE;
 break;
 case KW_EOF: nexttoken();
   x=parse_factor();
   mustbeint(x);
   opcode(OP_EOF);
   x=TY_INT;
 break;
 case KW_LOF: nexttoken();
   x=parse_factor();
   mustbeint(x);
   opcode(OP_LOF);
   x=TY_INT;
 break;
 case KW_FREEFILE: nexttoken();
   opcode(OP_FREEFILE);
   x=TY_INT;
 break;
 default: x=TY_INT;doerror(GE_EXP,0,0);
 }
 return x;
} 

int parse_variable(VARSYM **vp)
{
 VARSYM *p;
 int x,ndim;
 ndim=0;
 if(vartype==TY_PLAIN) vartype=default_type[tokname[0]-'A'];
 skipblanks();
 if(*parseptr=='(') vartype+=TY_ARRAY; /* indicate array variable */
 /* Note: we have to parse the left parenthesis 'by hand' because
    we still have to use the tokname and vartype variables */
 p=look_var();
 if(!(p->type&TY_FUNC)) { /* is it not a defined function ? */
   p->type=vartype;
   nexttoken(); /* skip the variable identifier token */
   if(p->type & TY_ARRAY) { /* is it an array? */
     do {
       nexttoken(); /* skip ( at first pass, comma at subsequent passes*/
       x=expr1();
       mustbeint(x);
       ndim++;
     } while(thistoken==TOK_COMMA);
     require(TOK_RPAR);
   }
 } else if(p==funcvar) nexttoken();
 *vp=p;
 x=p->type;
 if(p==funcvar)x=x&127;
 if((x&7)==TY_LONG) x=(x&~7)|TY_INT;
   else if((x&7)==TY_SINGLE)x=(x&~7)|TY_DOUBLE;
 return x+256*ndim;
}

void parse_defset(int k)
{ 
  char c,d;
  parseptr--;
  do{
  parseptr++;
  skipblanks();
  c=*parseptr++;
  if(!isalpha(c)) doerror(GE_SYN,0,0);
  if(islower(c))c=toupper(c);
  skipblanks();
  if(*parseptr=='-') {
    parseptr++;
    skipblanks();
    d=*parseptr++;
    if(!isalpha(d)) doerror(GE_SYN,0,0);
    if(islower(d)) d=toupper(d);
    skipblanks();
  } else d=c;
  while(d>=c) {
    default_type[d-'A']=k;
    d--;
  }
  } while(*parseptr==',');
  nexttoken();
}

void resolve_label(void)
     /* Resolve any forward references to label */
{
  LABSYM *lp;
  CELLTYPE *p,*q;
  lp=look_label();
  if(lp->flags&2) doerror(GE_DUP,0,0);
  p=lp->contents;
  while(p) {
   q=p;
   p=p->p;
   q->p=codeptr;
  }
  lp->contents=codeptr;
  lp->flags |= 2;
}

void use_label(void)
     /* Parse label and compile reference to it. */
{
 LABSYM * lp;
 if(thistoken != TOK_IDENT && thistoken != TOK_INT) doerror(GE_SYN,0,0);
 lp=look_label();
 if(lp->flags&2) { /* is label resolved? */
   opcptr(lp->contents);
 } else {
   opcptr(lp->contents);
   lp->contents=codeptr-1; /* link reference to unresolved label */
 }
 nexttoken();
}

void put_variable(int a,VARSYM *vp)
{
  if(vp==funcvar) { /* assign to the function return variable? */
    if(a==TY_INT) {
      opcode(OP_IPUTL);opcint(5+parm_slots);
    } else if(a==TY_DOUBLE) {
      opcode(OP_FPUTL);opcint(fparm_slots);
    } else if(a==TY_STR) {
      opcode(OP_SPUTL2);opcint(5+parm_slots);
    } else doerror(GE_TYP,0,0);
  } else {
    if(a==TY_INT) {
      if(vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
	opcode(OP_IPUT);
	opcptr(&(vp->contents.l));
      } else if(vp->contents.l>=2) {
	opcode(OP_IPUTR);
	opcint(parm_slots-vp->contents.l+6);
      } else if (vp->contents.l<=-2) {
	opcode(OP_IPUTL);opcint(vp->contents.l+1);
      } else if (vp->contents.l==0) {
	vp->contents.l=-2-(ilocals++);
	opcode(OP_IPUTL);opcint(vp->contents.l+1);
	lppush(vp);
      }
    } else if(a==TY_STR) {
      if(vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
	opcode(OP_SPUT);
	opcptr(&(vp->contents.p));
      } else if(vp->contents.l>=2) {
	opcode(OP_SPUTR);
	opcint(parm_slots-vp->contents.l+6);
      } else if (vp->contents.l<=-2) {
	opcode(OP_SPUTL);opcint(vp->contents.l+1);
      } else if (vp->contents.l==0) {
	vp->contents.l=-2-(slocals++);
	opcode(OP_SPUTL);opcint(vp->contents.l+1);
	lppush(vp);
      }
    } else if(a==TY_DOUBLE) {
      if(vp->contents.l==1||(vp->contents.l==0&&(insub==0||insub==3))) {
	opcode(OP_FPUT);
	opcptr(&(vp->contents.d));
      } else if(vp->contents.l>=2) {
	opcode(OP_FPUTR);
	opcint(parm_slots-vp->contents.l+6);
      } else if (vp->contents.l<=-2) {
	opcode(OP_FPUTL);opcint(vp->contents.l+1);
      } else if (vp->contents.l==0) {
	vp->contents.l=-2-(flocals++);
	opcode(OP_FPUTL);opcint(vp->contents.l+1);
	lppush(vp);
      }
    } else if((a&255)==TY_INT+TY_ARRAY) {
      bopcode(OP_IPUTA,3);
      opcint(a>>8);
      opcptr(&(vp->contents.p));
    } else if((a&255)==TY_STR+TY_ARRAY) {
      bopcode(OP_SPUTA,3);
      opcint(a>>8);
      opcptr(&(vp->contents.p));
    } else if((a&255)==TY_DOUBLE+TY_ARRAY) {
      bopcode(OP_FPUTA,3);
      opcint(a>>8);
      opcptr(&(vp->contents.p));
    }else doerror(GE_TYP,0,0);
    if(vp->type&128) doerror(GE_TYP,0,0);
  }
}

VARSYM * parse_decl(int x)
/* parse subroutine  or function  declaration 
   x: bit 0 (set for function), bit 1 (set if parameter adm must be built)
   bit 2 (set for DEFFN function)*/
{
      VARSYM * vp,*vp2;
      CELLTYPE *p1,*p2;
      int f=0,np=0;
      skipblanks();
      if(thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
      if((x&1) && vartype==TY_PLAIN) vartype=default_type[tokname[0]-'A'];
      if((x&1) && *parseptr=='(') vartype+=64;
      if(!(x&1)) {
        if (vartype!=TY_PLAIN) doerror(GE_SYN,0,0);
        vartype=TY_VOID;
      }
      vartype+=TY_FUNC;  
      vp=look_var();
      vp->type=vartype;
      if(x==3||x==7) { /* prepare function return variable */
	vartype=vartype&7;
	funcvar=look_var();
	if(funcvar->type==0) funcvar->type=vartype;
      }
      if(!vp->contents.p) { /* no subroutine descriptor present */
        f=1;
        bopcode(OP_JMP,128);
        p1=codeptr;
        opcptr(0);          /* jump past subroutine descriptor */
        vp->contents.p=p2=codeptr;
        if(x&4)opcint(0); else opcint(512); /* number of parameters */
        opcptr(0); /* pointer to subroutine code */
      } else p2=vp->contents.p; 
      nexttoken();
      if(thistoken==TOK_LPAR) {
	do{
	  nexttoken();
	  if(thistoken!=TOK_IDENT) doerror(GE_SYN,0,0);
          if(vartype==TY_PLAIN) vartype=default_type[tokname[0]-'A'];
	  if(f) {
            (p2->l)++; /* increase number of parameters */
	    opcint(vartype); /* add type to descriptor */ 
          } else { 
	    if(p2[2+np].l!=vartype) doerror(GE_PMM,0,0);
	    /* check type against descriptor */
	  }
          np++; /* increase number of parameters */
	  if(x&2) { /* we actually define the subroutine, so we have
		       to mark the symbol in the symbol table */
	    vp2=look_var();
	    vp2->type=vartype;
	    if(vp2->contents.l) doerror(GE_PMM,0,0);
	    if(vartype==TY_DOUBLE||vartype==TY_SINGLE) {
	      parm_slots+=1;
	      fparm_slots+=1;
	    } else {
	      parm_slots+=2;
	    }
	    vp2->contents.l=parm_slots+1;
	    lppush(vp2);      
	  }
          nexttoken();
        } while(thistoken==TOK_COMMA); 
	require(TOK_RPAR);
      }
      if(f) p1->p=codeptr;
      if((p2->l&255) != np) doerror(GE_PMM,0,0);
      return vp;
}

void close_sub(void)
     /* close the pending subroutine and restore symbol table */
{
  CELLTYPE t;
  int np;
  VARSYM *vp;
  bopcode(OP_UNLRET,3);
  opcint(parm_slots);
  opcint(fparm_slots);
  linkptr->l=ilocals;
  linkptr[1].l=slocals;
  linkptr[2].l=flocals;
  insub=0;
  ipop(np);
  while(lpptr) {
    lppop(vp);
    vp->contents.l=0;
  }
  funcvar=0;
  parm_slots=0;
  fparm_slots=0;
  ilocals=flocals=slocals=0;
}

void parse_actualparms(CELLTYPE *p1)
     /* parse the parameters of a function/subroutine call */
{
  VARSYM *vp;
  int np,i,x,pt;
  np=p1[0].l&255;
  for(i=0;i<np;i++) {
    pt=p1[2+i].l;
    skipblanks();
    if((p1[0].l&512) && thistoken==TOK_IDENT && (*parseptr==',' 
	|| *parseptr==')'||
	*parseptr==0 || *parseptr=='\n' || *parseptr==':' ||*parseptr=='\'')) {
      /* it's a variable */
      parse_variable(&vp);
      if(vp->type!=pt) doerror(GE_PMM,0,0);
      if(pt==TY_INT || pt==TY_LONG || pt==TY_STR) {
	opcode(OP_ILIT);opcint(0); /* assign dummy stack slot */
      } else opcode(OP_FZERO);
      if(vp->contents.l==0||vp->contents.l==1) { 
	/* reference to global variable */
	opcode(OP_SLIT);opcptr(&(vp->contents.p));
      } else if(vp->contents.l>=2) { /* referemce to paramter */
	opcode(OP_PARMREF);opcint(parm_slots-vp->contents.l+6);
      } else if(vp->contents.l<=-2) {
	if(pt==TY_INT || pt==TY_LONG) {
	  opcode(OP_ILOCREF);opcint(vp->contents.l+1);
	} else if(pt==TY_STR) {
	  opcode(OP_SLOCREF);opcint(vp->contents.l+1);
	} else {
	  opcode(OP_FLOCREF);opcint(vp->contents.l+1);
	}
      }
    } else { /* it's an expression, so pass by value */
      x=expr1();
      if(pt==TY_INT || pt==TY_LONG) {
	mustbeint(x);
	opcode(OP_REFSTK);
      } else if (pt==TY_SINGLE || pt==TY_DOUBLE) {
	mustbeflt(x);
	opcode(OP_REFFP);
      } else { /* is string */
	mustbestr(x);
	opcode(OP_REFSTK);
      }
    }	
    if(i!=np-1) { /* is not last parameter */
      require(TOK_COMMA);
    }
  }
}

void use_sub(CELLTYPE *p1)
{
  if(p1[0].l & 256) { /* already resolved ? */
    opcptr(p1[1].p);
  } else {
    opcptr(p1[1].p);
    p1[1].p=codeptr-1;
  }
}

void resolve_sub(CELLTYPE *p1)
{
  CELLTYPE *p,*q;
  if(p1[0].l & 256) doerror(GE_DUP,0,0);
  p1[0].l+=256;
  p=p1[1].p;
  while(p) {
   q=p;
   p=p->p; 	
   q->p=codeptr;
  }	  
  p1[1].p=codeptr;
}
